/* mlav.c -- routines that read a "refine" output file */

#include "util.h"
#include "mlav.h"
#include "seq.h"

#define BUF_SIZE 500000
#define MAX_SEQ 200000

static Mseq static_mseq[MAX_SEQ];

/* read the "refine"-output file to get the sequence names and the alignment
*  rows (alternating positive and negative numbers)
*/
static int read_align(char *filename, Mseq *mseq)
{
	FILE *fp;
	char buf[BUF_SIZE], *p, *q;
	int nseq, b, e, i;
	Block_ptr last, new;
	
	fp = ckopen(filename, "r");
	nseq = 0;
	while (fgets(buf, BUF_SIZE, fp)) {
		if (!strchr(buf, '\n'))
			fatal("raise BUFSIZE in read_align()");
		if (strncmp(buf, "s {", 3) == 0) {
			while (fgets(buf, BUF_SIZE, fp) && buf[0] != '}') {
				if ((p = strchr(buf, '"')) == NULL)
					fatalf("start of name", buf);
				if ((q = strchr(p+1, '"')) == NULL)
					fatalf("end of name", buf);
				*q = '\0';
				mseq[nseq].name = copy_string(p+1);
				if (sscanf(q+1, "%d %d", &b, &e) != 2)
					fatalf("positions: %s", buf);
				mseq[nseq].cur = mseq[nseq].beg = b;
				mseq[nseq].end = e;
				if (++nseq >= MAX_SEQ)
					fatal("Too many sequences.");
			}
		} else if (strncmp(buf, "a {", 3) == 0) {
			fgets(buf, BUF_SIZE, fp);  /* score */
			fgets(buf, BUF_SIZE, fp);  /* start points */
			fgets(buf, BUF_SIZE, fp);  /* end points */
			for (i=0; fgets(buf,BUF_SIZE,fp) && buf[0]!='}'; ++i) {
				if (strncmp(buf, " l ", 3))
					fatalf("line: %s", buf);
				last = NULL;
				for (p = buf + 3; sscanf(p, "%d", &b) == 1; ) {
					new = ckalloc(sizeof(Block));
					new->size = b;
					new->next = NULL;
					if (last == NULL)
						mseq[i].scan = 
						  mseq[i].block = new;
					else
						last->next = new;
					last = new;
					while (*p == ' ')
						++p;
					while (!strchr(" \n\0", *p))
						++p;
				}
			}
			if (i != nseq) {
				fprintf(stderr, "i = %d, nseq = %d: ",
					i, nseq);
				fatal("i != nseq");
			}
		}
	}
	fclose(fp);
	return nseq;
}

int read_mseqs(char *filename, Mseq **mseq)
{
	int i, len, nseq, revcomp;
	SEQ *sf = NULL;
	char root[500], oldroot[500], buf[500], *s;
	Mseq *seq = static_mseq;

	nseq = read_align(filename, static_mseq);

	buf[0] = oldroot[0] = '\0';
	for (i = 0; i < nseq; ++i) {
		/* Is this the previous sequence or its reverse complement? */
		if (i > 0) {
			if (same_string(seq[i].name, seq[i-1].name)) {
				/* this is the current sequence */
				seq[i].s = seq[i-1].s;
				continue;
			}
			strcpy(buf, seq[i-1].name);
			strcat(buf, "-");
			if (same_string(seq[i].name, buf)) {
				/* reverse complement of the current sequence */
				seq_revcomp_inplace(sf);
				seq[i].s = copy_string(SEQ_CHARS(sf));
				continue;
			}
		}
		/* see if we want a later contig in the same sequence file */
		strcpy(root, seq[i].name);
		len = strlen(root);
		if ((revcomp = (root[len-1] == '-')) == 1)
			root[len-1] = '\0';
		if ((s = strchr(root, ':')) != NULL)
			*s++ = '\0';
		if (strcmp(root, oldroot)) {
			/* no -- it is in a new sequence file */
			if (sf)
				seq_close(sf);
			sf = seq_get(root);
		}
		strcpy(oldroot, root);
		if (s != NULL) {	/* contig specified */
			len = strlen(s);
			while (strncmp(s, SEQ_HEAD(sf)+1, len) != 0)
				if (!seq_read(sf))
					fatalf("Cannot find %s", seq[i].name);
		}
		/* now we have the right sequence */
		if (revcomp)
			seq_revcomp_inplace(sf);
		seq[i].s = copy_string(SEQ_CHARS(sf));
	}
	if (sf)
		seq_close(sf);
	*mseq = static_mseq;
	return nseq;
}

/* strip off trailing dash and characters starting with the first ":" */
static void root_name(char *name)
{
	char *s;
	int len = strlen(name);

	if (name[len-1] == '-')
		name[len-1] = '\0';
	if ((s = strchr(name, ':')) != NULL)
		*s = '\0';
}

char **malign_matrix(char *filename, int *nrow_ptr, int *ncol_ptr,
  char ***seq_name_ptr, char unaligned) {
	Mseq *mseq;
	int nseq, i, j, size, nrow, ncol, row;
	Block_ptr bp;
	char **A, **seq_name, *s, *t;

	nseq = read_mseqs(filename, &mseq);
	/* Determine the number of columns. */
	for (ncol = 0, bp = mseq[0].block; bp; bp = bp->next)
		ncol += ((size = bp->size) > 0 ? size : -size);

	/* Determine the number of rows (species), where the root name tells
	*  the species.
	*/
	root_name(mseq[0].name);
	for (nrow = i = 1; i < nseq; ++i) {
		root_name(mseq[i].name);
		if (!same_string(mseq[i].name, mseq[i-1].name))
			++nrow;
	}

	/* Allocate the matrix and fill it with dashes. */
	A = ckalloc(nrow * sizeof(char *));
	A[0] = ckalloc (nrow * ncol * sizeof(char));
	for (i = 1; i < nrow; ++i)
		A[i] = A[i-1] + ncol;
	for (i = 0; i < nrow; ++i)
		for (j = 0; j < ncol; ++j)
			A[i][j] = unaligned;

	/* Fill in the nucleotides */
	seq_name = ckalloc(nrow * sizeof(char *));
	for (i = 0, row = -1; i < nseq; ++i) {
		if (i == 0 || strcmp(mseq[i].name,mseq[i-1].name))
			seq_name[++row] = copy_string(mseq[i].name);
		t = mseq[i].s + mseq[i].beg - 1; /* first aligned position */
		for (s = A[row], bp = mseq[i].block; bp; bp = bp->next)
			if ((size = bp->size) > 0)
				for (j = 0; j < size; ++j) {
					if (*s != unaligned)
					  fatal("not a single-cov alignment.");
					*s++ = *t++;
				}
			else if (bp == mseq[i].block)
				s -= size;
			else if (bp->next != NULL)
				while (size++ < 0)
					*s++ = '-';
	}

	*nrow_ptr = nrow;
	*ncol_ptr = ncol;
	*seq_name_ptr = seq_name;
	return A;
}
