/*
* OandO.c - predict an order and orientation for contigs from a blastz output
*
* The command syntax is "OandO alignment_file".
*
* The second sequence in the blastz output is assumed to have been divided
* into "BACs", each of which contains one or more "contigs".  The FastA
* header line for a BAC includes a 1-word "name" (as extracted by the
* procedure fasta_name), and subsequent contigs have FastA header lines
* name.CN for N = 2, 3 ...  (Such contig headers are generated by the
* program "contigs" at positions where the sequence has a long run of N's.)
*/

static const char rcsid[] =
"$Id: OandO.c,v 1.8 2004/01/13 21:09:13 schwartz Exp $";

enum {
  MAX_CONTIGS=1*1000*1000,
  MAX_BACS=500*1000,
  BUF_SIZE=5000
};

#include "util.h"
#include "lav.h"

typedef struct bac {
	char *root;
	int first_contig, last_contig; // half-open on right
	int max_score, best_pos;
} BAC;

BAC B[MAX_BACS];

typedef struct cont {
	char *header, *name, *root;
	int nth, rc;
	int score, beg, end, span_from, span_to;
} Contig;

Contig C[MAX_CONTIGS];

void chkB(int n) {
    if (n<0) fatal("negative BAC index");
    if (n >= MAX_BACS) fatal("too many BACs");
}
void chkC(int n) {
    if (n<0) fatal("negative contig index");
    if (n >= MAX_CONTIGS) fatal("too many contigs");
}

char *rc_to_dash(char *s)
{
	char *p = strstr(s, " (reverse complement)");
	if (p) strcpy(p, "-");
	return s;
}

/* order contigs within a BAC by the start position of their highest scoring
*  alignment.
*/
void order(int first_contig, int last_contig, int threshold) {
	int i, j, huge;
	char buf[BUF_SIZE];

	/* first come the contigs having an alignment */
	huge = 0;
	for (i = first_contig; i < last_contig; ++i)
		huge = MAX(huge, C[i].beg);
	++huge;
	for ( ; ; ) {
		for (i = first_contig; i < last_contig; ++i)
			if (C[i].score > 0)
				break;
		for (j = i+1; j < last_contig; ++j)
			if (C[j].score > 0 && C[j].beg < C[i].beg)
				i = j;
		if (i >= last_contig)
			break;
		if (C[i].beg >= huge)
			break;
		strcpy(buf, C[i].header);
		rc_to_dash(buf);
		printf("%d-%d\t%5.1f\t %s\n", C[i].span_from, C[i].span_to,
		  (float)C[i].score/(float)threshold, buf);
		C[i].beg = huge;
	}
	/* end with the contigs having no alignment, in their initial order */
	for (i = first_contig; i < last_contig; ++i)
		if (C[i].score == 0)
			printf("%s\n", rc_to_dash(C[i].header));
}

char *root_name(char *name)
{
	char *s, *t;

	// Look for ".C" followd by an integer...
	s = strstr(name, ".C");
	if (s == 0 || s[2] == 0) return "";
	strtol(s+2, &t, 10);
	if (*t != 0) return "";

	// ...and return the prefix.
	t = copy_string(name);
	t[s-name] = 0;
	return t;

/*
// This fails for >C1 >C10 >C111
	char buf[5000];
	strcpy(buf, name);
	s = buf + strlen(buf) - 1;
	if (isdigit(*s)) {
		while (isdigit(*s))
			--s;
		if (s[0] == 'C' && s[-1] == '.')
			s[-1] = '\0';
	}
	return copy_string(buf);
*/
}

int main(int argc, char **argv)
{
	FILE *ap;
	char buf[500], *status, *p, *q;
	int i, j, val, huge, nc, nb, threshold = 0, from, to;

	argv0 = "OandO";
	if (argc != 2)
		fatal("arg = blastz-output");

	/* read contigs and their highest scoring alignment */
	ap = ckopen(argv[1], "r");

	nc = -1;
	for ( ; ; ) {
		if  ((status = fgets(buf, 500, ap)) == NULL ||
		     strstr(buf, "s {")  // XXX -- not strstr(buf, "#:lav")
		) {
			int rc, nth;

			// Equate the previously seen
			// contigs and its reverse complement
			// and take the better of the two.
			// XXX - assumes they come in this order.
			if (nc >= 1 &&
			    C[nc].nth == C[nc-1].nth &&
			    C[nc].rc == 1 && C[nc-1].rc == 0)
			{
//fprintf(stderr, "equating {%s} {%s}\n", C[nc-1].header,C[nc].header);
				if (C[nc-1].score < C[nc].score) {
					free(C[nc-1].header);
					C[nc-1] = C[nc];
					C[nc].beg = C[nc].end =
					  C[nc].span_from = C[nc].span_to = 0;
				}
				--nc;
			}

			if (status == NULL)
				break;

			++nc;
			chkC(nc);
			fgets(buf, 500, ap);
			if (2 != sscanf(buf, " %*s %*d %*d %d %d", &rc, &nth))
			    fatalf("sscanf {%s}",buf);
			fgets(buf, 500, ap);
			if (2 != sscanf(buf, " %*s %*d %*d %d %d", &rc, &nth))
			    fatalf("sscanf {%s}",buf);
			C[nc].score = C[nc].span_from =  0;
			C[nc].nth = nth;
			C[nc].rc = rc;
//fprintf(stderr, "new contig[%d]=%d\n", nc, nth);
		}
		if (strstr(buf, "K = ") && (p = strstr(buf, "L = ")) != NULL)
			threshold = atoi(p+4);
		if (strncmp(buf, "h {", 3) == 0) {
			/* get the next contig's name */
			fgets(buf, 500, ap);
			fgets(buf, 500, ap);
			if ((p = strchr(buf+2, '"')) == NULL)
				fatalf("quotation: %s", buf);
			++p;
			if ((q = strchr(p, '"')) != NULL)
				*q = '\0';
			chkC(nc);
			C[nc].header = copy_string(p);
			C[nc].name = fasta_name(p);
			C[nc].root = root_name(C[nc].name);
		} else if (strncmp(buf,"a {", 3) == 0) {
			/* get alignment score and end-points in first seq */
			fgets(buf, 500, ap);
			if (sscanf(buf, "  s %d", &val) != 1)
				fatalf("score: %s", buf);
			fgets(buf, 500, ap);
			if (sscanf(buf, "  b %d", &from) != 1)
				fatalf("begin: %s", buf);
			fgets(buf, 500, ap);
			if (sscanf(buf, "  e %d", &to) != 1)
				fatalf("end: %s", buf);
			/* check for new maximum score */
			chkC(nc);
			if (val > C[nc].score) {
				C[nc].score = val;
				C[nc].beg = from;
				C[nc].end = to;
			}
			/* update the span of this contig in this direction */
			if (C[nc].span_from == 0) {
				/* first alignment of this contig */
				C[nc].span_from = from;
				C[nc].span_to = to;
			} else {
				C[nc].span_from = MIN(C[nc].span_from, from);
				C[nc].span_to = MAX(C[nc].span_to, to);
			}
			
		}
	}
	fclose(ap);
//fprintf(stderr, "nc=%d\n", nc);

	if (threshold == 0)
		fatal("Cannot determine threshold.");

	/* group contigs into BACs according to "root name" */
	B[0].root = C[0].root;
	B[0].max_score = C[0].score;
	B[0].best_pos = C[0].beg;
	B[0].first_contig = 0;
	B[0].last_contig = 0;
	nb = 0;
	for (i = 0; i <= nc; ++i) {
		chkC(i); chkB(nb);
		if (same_string(C[i].root, B[nb].root)) {
			B[nb].last_contig = i+1;
			if (C[i].score > B[nb].max_score) {
				B[nb].best_pos = C[i].beg;
				B[nb].max_score = C[i].score;
			}
		} else {
			++nb;
			chkB(nb);
			B[nb].root = C[i].root;
			B[nb].max_score = C[i].score;
			B[nb].best_pos = C[i].beg;
			B[nb].first_contig = i;
			B[nb].last_contig = i+1;
		}
	}
//fprintf(stderr, "nb=%d\n", nb);

	/* order BACs by the start position of highest scoring alignment. */
	for (i = huge = 0; i <= nb; ++i)
		huge = MAX(huge, B[i].best_pos);
	++huge;
	for ( ; ; ) {
		for (i = 0; i <= nb; ++i)
			if (B[i].max_score > 0)
				break;
		for (j = i+1; j <= nb; ++j)
			if (B[j].max_score > 0 && B[j].best_pos < B[i].best_pos)
				i = j;
		if (B[i].best_pos >= huge)
			break;
		order(B[i].first_contig, B[i].last_contig, threshold);
		B[i].best_pos = huge;
	}
	/* end with the BACs having no alignment, in their initial order */
	for (i = 0; i <= nb; ++i)
		if (B[i].max_score == 0)
			for (j = B[i].first_contig; j < B[i].last_contig; ++j)
				printf("%s\n", rc_to_dash(C[j].header));

	return 0;
}
