/* analyze_exons.c
*
*  syntax:
*	analyze_exons [-d] exons-file blastz-output
*
*  The program attempts to use the alignments to map each position in the
*  "exons" file into the corresponding position in the second sequence.
*  For each coding-region specification (i.e., line beginning with "+")
*  the first and last codon are displayed, and for each intron the first and
*  last two nucleotides are shown.  These tri- and di-nucleotides are shown
*  for the first sequence and, if possible, for the second sequence.  If the
*  coding region is specified and the "-d" flag is not used, then the putative
*  coding region is printed (for both sequences, if possible).  For example,
*  if the exons file contains:

< 27591 30475 L44L
+ 27641 30438
27591 27661
27932 28054
29660 29727
29918 30023
30436 30475

*  then the program might generate:

< L44L: 27591-30475     35159-37479
  CDS:  27641-30438     35202-37442     TTA-CAT TTA-CAT
   5:   27591-27661     35159-35222     CT-AC   CT-AC
   4:   27932-28054     35586-35708     CT-AC   CT-AC
   3:   29660-29727     36581-36648     CT-AC   CT-AC
   2:   29918-30023     36976-37081     CT-AC   CT-AC
   1:   30436-30475     37440-37479

>L44L, putative CDS for sequence 1 (321 bp)
ATGGTTAACGTCCCTAAAACCCGCCGGACTTTCTGTAAGAAGTGTGGCAA
GCACCAACCCCATAAAGTGACACAGTACAAGAAGGGCAAGGATTCTCTGT
  ....

*  This output shows that the L44L coding region in both sequences begins with
*  ATG (the reverse complement of CAT) and ends with TAA.  Similarly all four
*  introns conform to the normal GT-AG splicing consensus.
*
*  When the second sequence file is split into contigs, the output is somewhat
*  more complicated.  It begins with an enumeration of the contigs (using their
*  FastA header lines), and positions in the second sequence are identified
*  as  i:j which denotes position j in contig i.  For instance, the index might
*  include:

Index of fragments of the second sequence:
	1: >Contig101
	2: >Contig102

* and the listing of exon positions might contain:

< L44L:	27591-30475	1:35159-2:1479
  CDS:	27641-30438	1:35202-2:1442	TTA-CAT	TTA-CAT
   5:	27591-27661	1:35159-1:35222	CT-AC	CT-AC
   4:	27932-28054	1:35586-1:35708	CT-AC	CT-AC
   3:	29660-29727	2:581-2:648	CT-AC	CT-AC
   2:	29918-30023	2:976-2:1081	CT-AC	CT-AC
   1:	30436-30475	2:1440-2:1479

* Thus the left-most CDS position (end of the stop codon) aligns with position
* 35202 in fragment 1, which has FastA header line ">Contig101".
*/

static const char rcsid[]=
"$Id: analyze_exons.c,v 1.6 2002/08/23 06:12:29 schwartz Exp $";

#include "util.h"
#include "seq.h"
#include "exons.h"
#include "lav.h"

enum {
	BUF_SIZE=1000, MAX_RNA=60*1000*1000, MAX_OLIGO=4
};

#define rc(x) dna_cmpl(toupper(x))

void flip_seq(uchar *s)
{
	uchar temp, *t;

	for (t = s + strlen((char *)s) - 1; s < t; ++s, --t) {
		temp = rc(*s);
		*s = rc(*t);
		*t = temp;
	}
}

void print_seq(uchar *s, char *name, int which, int bp, int dir)
{
	uchar *start = s;

	printf("\n>%s, putative CDS for sequence %d (%d bp)\n",
		name, which, bp);
	if (dir == -1)
		flip_seq(s);
	for ( ; *s; ++s) {
		putchar((char)*s);
		if ((s-start+1)%50 == 0)
			putchar('\n');
	}
	putchar('\n');
}

int get_fragments(char *seqname, uchar ***seq_addr, int **N_addr)
{
	uchar **seq;
	SEQ *sp;
	int nseq, i, *N;

	if ((sp = seq_get(seqname)) == NULL)
		fatalf("Couldn't get sequence '%s'.", seqname);
	for (nseq = 1; seq_read(sp); ++nseq)
		;
	seq_close(sp);

	*seq_addr = seq = ckallocz(nseq*sizeof(uchar *));
	*N_addr = N = ckallocz(nseq*sizeof(int));
	for (i = 0, sp = seq_get(seqname); i < nseq; ++i, seq_read(sp)) {
		seq[i] = (uchar *) copy_string((const char *)SEQ_CHARS(sp));
		N[i] = SEQ_LEN(sp);
	}

	seq_close(sp);
	return nseq;
}

void free_fragments(uchar **seq, int *N, int n)
{
	while (--n >= 0)
		free(seq[n]);
	free(seq);
	free(N);
}

void list_contigs(FILE *fp, int nlav)
{
	char buf[BUF_SIZE], cname[BUF_SIZE];
	int i;

	printf("Index of fragments of the second sequence:\n");
	buf[0] = '\0';
	for (i = 1; i <= nlav; ++i)
		printf("\t%d: %s\n", i,
			lav_contigname(fp, buf, BUF_SIZE, cname));
}

/* oligo -- extract a few nucleotides from a sequence */
char *oligo(uchar *seq, int N, int base, int from, int to, char *target)
{
	int length, i, fail;

	if ((length = to - from + 1) > MAX_OLIGO)
		fatal("oligo is too long");

	fail = (base <= 0 || base - from <= 0 || base + to > N);

	for (i = 0; i < length; ++i)
		target[i] = (fail ? '?' : (char)seq[base+from+i]);
	target[length] = '\0';
	return target;
}

void ckcpy(uchar *s, uchar c, uchar *start, char *g)
{
	*s = c;
	if ((s-start) >= MAX_RNA-1)
		fatalf("Gene %s is too long.", g);
}

void print_CDS(int gene, int n_exons, int *beg1, int *end1, int *dir,
  char **gene_name, uchar *seq1, int *beg2, int *end2, int *nlav_b,
  int *nlav_e, uchar **seq2)
{
	uchar *newseq = ckallocz(MAX_RNA);
	uchar *s;
	int n, CDS_beg1, CDS_end1, from, to, CDS_beg2, CDS_end2, CDS_nlav_b,
		CDS_nlav_e;
	char *g = gene_name[gene];

	s = newseq;
	n = gene + 1;
	CDS_beg1 = beg1[n];
	CDS_end1 = end1[n];
	if (CDS_beg1 <= 0 || CDS_end1 <= 0)
		goto out;
	for (++n; n < n_exons && gene_name[n] == NULL; ++n) {
		from = MAX(beg1[n], CDS_beg1);
		to = MIN(end1[n], CDS_end1);
//fprintf(stderr, "exon %d from %d to %d\n", n, from, to);
		while (from <= to)
			ckcpy(s++, seq1[from++], newseq, g);
	}
	*s = '\0';
	print_seq(newseq, g, 1, s-newseq, dir[gene]);

	/* Print putative coding sequence for the second sequence file.
	*  We require that the two endpoints of each coding exon are mapped
	*  into the same fragment in the second sequence.
	*/
	n = gene + 1;
	CDS_beg2 = beg2[n];
	CDS_end2 = end2[n];
	CDS_nlav_b = nlav_b[n];
	CDS_nlav_e = nlav_e[n];
	if (CDS_beg2 <= 0 || CDS_end2 <= 0)
		goto out;
	s = newseq;
	/* search for first coding exon */
	for (++n; n < n_exons && gene_name[n] == NULL; ++n)
		if (beg1[n] <= CDS_beg1 && CDS_beg1 <= end1[n])
			break;
	if (n == n_exons || gene_name[n] != NULL)
		goto out;
	if (CDS_end1 <= end1[n]) {	/* just one coding exon */
		if (CDS_nlav_b != CDS_nlav_e)
			goto out;
		for (from = CDS_beg2-1; from < CDS_end2; ++from)
			ckcpy(s++, seq2[CDS_nlav_b][from], newseq, g);
		*s = '\0';
		print_seq(newseq, g, 2, s-newseq, dir[gene]);
		goto out;
	}
	if (CDS_nlav_b != nlav_e[n] || end2[n] == 0)
		goto out;
	for (from = CDS_beg2-1; from < end2[n]; ++from) {
		ckcpy(s++, seq2[CDS_nlav_b][from], newseq, g); // XXX
	}

	/* handle internal coding exons */
	for (++n;
	   n < n_exons && gene_name[n] == NULL && end1[n] < CDS_end1; ++n) {
		if (nlav_b[n] != nlav_e[n] || beg2[n] == 0 || end2[n] == 0)
			goto out;
		for (from = beg2[n]-1; from < end2[n]; ++from)
			ckcpy(s++, seq2[nlav_b[n]][from], newseq, g);
	}

	/* handle last coding exon */
	if (n == n_exons || gene_name[n] != NULL)
		goto out;
	if (nlav_b[n] != CDS_nlav_e || beg2[n] == 0)
		goto out;
	for (from = beg2[n]-1; from < CDS_end2; ++from)
		ckcpy(s++, seq2[nlav_b[n]][from], newseq, g); // XXX
	
	*s = '\0';
	print_seq(newseq, g, 2, s-newseq, dir[gene]);
out:
	ckfree(newseq);
}

void print2(int i, int j)
{
	putchar('\t');
	if (i > 0)
		printf("%d", i);
	else
		printf("??");
	putchar('-');
	if (j > 0)
		printf("%d", j);
	else
		printf("??");
}

void print4(int m, int i, int n, int j)
{
	putchar('\t');
	if (i > 0)
		printf("%d:%d", m+1, i);
	else
		printf("??");
	putchar('-');
	if (j > 0)
		printf("%d:%d", n+1, j);
	else
		printf("??");
}


int main(int argc, char **argv)
{
	FILE *fp;
	int i, j, n, n_exons, from1, from2, to1, to2, incr, next_gene, gene,
		N1, dna = 1, nlav, nseq2,
		*beg1,	/* exon i begins at position beg1[i] seq1	*/
		*end1,  /* exon i ends at position end1[i] seq1		*/
		*beg2,	/* exon i begins at position beg2[i] seq2	*/
		*end2,  /* exon i ends at position end2[i] seq2		*/
		*nlav_b, /* exon i begins on nth fragment in lav */
		*nlav_e, /* exon i ends on nth fragment in lav */
		*nseq_b, /* exon i begins on fragment nseq_b[i] in seq2	*/
		*nseq_e, /* exon i ends on fragment nseq_e[i] in seq2	*/
		*dir,	/* dir[n] = 1 means left-to-right transcription	*/
		*N2;	/* fragment k in seq2 has length N2[k]		*/
	char buf[BUF_SIZE], buf2[BUF_SIZE], *g, **gene_name, **exon_name,
		o1[MAX_OLIGO+1], o2[MAX_OLIGO+1];
	SEQ *sp1;
	uchar *seq1,	/* seq1 is seq1[1], seq1[2], ...		*/
		**seq2;	/* fragment k in seq2 is seq2[k][0], seq2[k][1],...  */
			/* ??? k = 0,1,...; second index = 0,1,... */
	int nth, rc;

	argv0 = "analyze_exons";

	if (argc == 4 && strcmp(argv[1], "-d") == 0)
		dna = 0;
	if (argc != 3)
		fatal("args = [-d] exons-file blastz-output");

	n_exons = get_exons(argv[argc-2], &beg1, &end1, &dir, &gene_name,
	  &exon_name, 0);
	if (n_exons == 0)
		fatal("no genes or exons were specified");

	fp = alignopen(argv[argc-1]);

	/* get the first sequence */
	
	buf[0] = '\0';
	if (lav_seqname(1 , fp, buf, BUF_SIZE, buf2, &from1, &to1) == 0) {
	    fprintf(stdout, "# no alignments\n");
	    exit(0);
	}

	if (from1 != 1)
		fatal("Not implemented for sequence subrange.");
	if ((sp1 = seq_get(buf2)) == NULL)
		fatalf("Couldn't get sequence '%s'.", buf);
	N1 = SEQ_LEN(sp1);
	seq1 = SEQ_CHARS(sp1) - 1;

	/* another check for improper "exons" file */
	for (n = 0; n < n_exons; ++n)
		if (end1[n] > N1)
		  fatalf("position %d in exons file; sequence length is %d\n",
			end1[n], N1);

	beg2 = ckallocz(n_exons*sizeof(int));
	end2 = ckallocz(n_exons*sizeof(int));
	nlav_b = ckallocz(n_exons*sizeof(int));
	nlav_e = ckallocz(n_exons*sizeof(int));
	nseq_b = ckallocz(n_exons*sizeof(int));
	nseq_e = ckallocz(n_exons*sizeof(int));
	for (n = 0; n < n_exons; ++n)
		beg2[n] = end2[n] = nlav_b[n] = nlav_e[n] = 0;

	/* map beg1 and end1 positions to nlav_b:beg2 and nlav_e:end2 */ 
	rewind(fp);
	nlav = 0;
	nth = 1;
	rc = 0;
	while (fgets(buf, BUF_SIZE, fp)) {
		if (sscanf(buf, "  l %d %d %d %d %*d",
		    &from1, &from2, &to1, &to2) == 4)
			for (n = 0; n < n_exons; ++n) {
				if (from1 <= beg1[n] && beg1[n] <= to1) {
					beg2[n] = from2 + (beg1[n] - from1);
					nlav_b[n] = nlav-1;
					nseq_b[n] = 2*(nth-1)+rc;
/*
fprintf(stderr, "n=%d beg2=%d nlav_b=%d nseq_b=%d ",
	    n, beg2[n], nlav_b[n], nseq_b[n]);
fprintf(stderr, "from1=%d beg1=%d to1=%d from2=%d\n",
	    from1, beg1[n], to1, from2);
*/
				}
				if (from1 <= end1[n] && end1[n] <= to1) {
					end2[n] = from2 + (end1[n] - from1);
					nlav_e[n] = nlav-1;
					nseq_e[n] = 2*(nth-1)+rc;
/*
fprintf(stderr, "n=%d end2=%d nlav_e=%d nseq_e=%d ",
	    n, end2[n], nlav_e[n], nseq_e[n]);
fprintf(stderr, "from1=%d end1=%d to1=%d from2=%d\n",
	    from1, end1[n], to1, end2);
*/
				}
			}
		else if (!strcmp(buf, "s {\n"))	{	// XXX -- was #:lav
		    int lo,hi;
		    fgets(buf, BUF_SIZE, fp);
		    fgets(buf, BUF_SIZE, fp);
		    n = lav_scan_s(buf, buf2, &lo, &hi, &rc, &nth);
//fprintf(stderr, "s -> %d %d\n", rc, nth);
		    if (n != 5) fatalf("bad s record (%s)\n", buf);
		    ++nlav;
		}
	}
	rewind(fp);
//fprintf(stderr, "%d stanzas\n", nlav);
	if (nlav > 1)
		list_contigs(fp, nlav);

	rewind(fp);
	buf[0] = '\0';
	(void) lav_seqname(2 , fp, buf, BUF_SIZE, buf2, &from2, &to2);
	if (from2 != 1)
		fatal("Not implemented for sequence subrange.");
	fclose(fp);
	nseq2 = get_fragments(buf2, &seq2, &N2);
//fprintf(stderr, "%d fragments\n", nseq2);

	if (1) {	/* always search reverse complements */
		uchar **tmp = ckallocz(nseq2*2*sizeof(uchar *));
		int *nn = ckallocz(nseq2*2*sizeof(int));
		int i;

		for (i = 0; i < nseq2; ++i) {
			tmp[2*i] = seq2[i];
			tmp[2*i+1] = (uchar *)copy_string((const char *)seq2[i]);
			nn[2*i+1] = nn[2*i] = N2[i];
			flip_seq(tmp[2*i+1]); /* assumes &to2 is always full length of seq */
		}
		free(seq2);
		free(N2);
		seq2 = tmp;
		N2 = nn;
	}
	//assert(nlav <= nseq2);

	/* generate output */
	for (n = 0; n < n_exons-1; n = next_gene) {
		if ((g = gene_name[n]) == NULL)
			fatal("Could not find next gene.");
		printf("\n____________________________");
		printf("______________________________\n\n");
		printf("%c %s:", dir[n]==1 ? '>' : '<', g);
		print2(beg1[n], end1[n]);
		if (nlav == 1)
			print2(beg2[n], end2[n]);
		else
			print4(nlav_b[n], beg2[n], nlav_e[n], end2[n]);
		putchar('\n');
		gene = n++;
		if (beg1[n]) {
			printf("  CDS:");
			print2(beg1[n], end1[n]);
			if (nlav == 1)
				print2(beg2[n], end2[n]);
			else
				print4(nlav_b[n], beg2[n], nlav_e[n], end2[n]);
			printf("\t%s-%s",
			    oligo(seq1, N1, beg1[n], 0, 2, o1),
			    oligo(seq1, N1, end1[n], -2, 0, o2));
			printf("\t%s-%s",
			    oligo(seq2[nseq_b[n]]-1, N2[nseq_b[n]],
				beg2[n], 0, 2, o1),
			    oligo(seq2[nseq_e[n]]-1, N2[nseq_e[n]],
				end2[n], -2, 0, o2));
		} else
			printf("  CDS not specified");
		putchar('\n');
		for (j = ++n; j < n_exons && gene_name[j] == NULL; ++j)
			;
		next_gene = j;
		if (dir[gene] == 1) {
			i = 1;
			incr = 1;
		} else {
			i = next_gene - n;	/* number of exons in gene */
			incr = -1;
		}
		for ( ; n < next_gene; ++n, i += incr) {
			printf("%4d:", i);
			print2(beg1[n], end1[n]);
			if (nlav == 1)
				print2(beg2[n], end2[n]);
			else
				print4(nlav_b[n], beg2[n], nlav_e[n], end2[n]);
			if (n < next_gene - 1) {
				printf("\t%s-%s",
				    oligo(seq1, N1, end1[n], 1, 2, o1),
				    oligo(seq1, N1, beg1[n+1], -2, -1, o2));
				printf("\t%s-%s",
				    oligo(seq2[nseq_e[n]]-1, N2[nseq_e[n]],
					end2[n], 1, 2, o1),
				    oligo(seq2[nseq_b[n+1]]-1, N2[nseq_b[n+1]],
					beg2[n+1], -2, -1, o2));
			}
			putchar('\n');
		}

		if (beg1[gene+1] > 0 && dna)	/* CDS was specified */
			print_CDS(gene, n_exons, beg1, end1, dir,
			    gene_name, seq1, beg2, end2, nseq_b, nseq_e, seq2);
	}
	
	free_exons(beg1, end1, dir, gene_name, exon_name, n_exons);
	seq_close(sp1);
	free_fragments(seq2, N2, nlav);
	free(beg2);
	free(end2);
	free(nlav_b);
	free(nlav_e);
	return 0;
}
