/*
* order_lav.c - order an alignment file according to an OandO output
*
* Syntax: order_lav OandO-output blastz-output
*
* We need to identify routines shared with single_cov (get_lines, free_lines).
*/

#include "util.h"
#include "lav.h"

#define RC " (reverse complement)"

enum {
  BUF_SIZE=500,
  MAX_CONTIG=500*1000
};

/* The input is stored as a linked list of arbitrary lines, with interspersed
alignments */
typedef struct line_rec {
	char *line;
	struct line_rec *next;
} Line, *Line_ptr;

typedef struct contig {
	char *h1, *name;
	char *s1, *s2;
	int s1lo, s1hi;
	int s2lo, s2hi, s2rc;
	Line_ptr list[4]; /* 0 = forward header, 1 = forward alignments,
			     2 = reverse header, 3 = reverse alignments */
	Line_ptr last[4];
} Contig;

Contig contig[MAX_CONTIG];
void chkC(int n) { if (n<0) fatal("negative congig index"); if (n >= MAX_CONTIG) fatal("too many contigs"); }


void save_line(char *buf, int i, int list)
{
	Line *new = ckalloc(sizeof(Line));
	new->line = copy_string(buf);
	new->next = NULL;
	chkC(i);
	if (contig[i].list[list] == NULL)
		contig[i].list[list] = new;
	else
		contig[i].last[list]->next = new;
	contig[i].last[list] = new;
//fprintf(stderr, "saving [%d][%d] <- [%s]\n", i, list, buf);
}

void unquotify(char *s)
{
	int l = strlen(s);
	if (s[l-1] == '"') s[l-1] = 0;
	if (s[l-2] == '-') s[l-2] = 0;
	if (s[0] == '"') memmove(s, s+1, l-1);
}

void unRCify(char *s)
{
	char *p = strstr(s, RC);
	if (p) *p = 0;
}


int get_contigs(FILE *fp)
{
	char buf[BUF_SIZE], buf1[BUF_SIZE], buf2[BUF_SIZE], name[BUF_SIZE];
	int rc = 0;
	int type = 0;
	int max = 0;
	int nth;

// we assume that the input consists of a contig 
// optionally going forward, then optinally going reverse
//   prev this  quit?
//   fwd, rev -> more
//   rev, rev -> done
//   fwd, fwd -> done
//   rev, fwd -> done

	while (fgets(buf, BUF_SIZE, fp)) {
		// identify tokens that come before 'a'
		// XXX - bad design to require a complete list here.
		if (same_prefix(buf, "d {"))
			type = 'd';
		else if (same_prefix(buf, "s {"))
			type = 's';
		else if (same_prefix(buf, "h {"))
			type = 'h';
		else if (same_prefix(buf, "#:lav"))
			type = ':';
		else if (same_prefix(buf, "a {"))
			type = 'a';

		if (type == ':') {
		}

		if (type == 's') {
			int n, s1lo, s1hi, s2lo, s2hi;
			char *s1;
			fgets(buf1, BUF_SIZE, fp); // "seq1"
			fgets(buf2, BUF_SIZE, fp); // "seq2"
			type = 0;

			n  = lav_scan_s(buf1, name, &s1lo, &s1hi, &rc, &nth);
			s1 = copy_string(name);
			n  = lav_scan_s(buf2, name, &s2lo, &s2hi, &rc, &nth);

			contig[nth].s1 = s1;
			contig[nth].s1lo = s1lo;
			contig[nth].s1hi = s1hi;
			unquotify(contig[nth].s1);
//fprintf(stderr, "s1=%s nth=%d\n", contig[nth].s1, nth);

			contig[nth].s2 = copy_string(name);
			contig[nth].s2lo = s2lo;
			contig[nth].s2hi = s2hi;
			contig[nth].s2rc = rc;
			unquotify(contig[nth].s2);
//fprintf(stderr, "s2=%s n=%d\n", contig[nth].s2, n);

			//save_line(buf,  nth, rc ? 2 : 0);
			//save_line(buf1, nth, rc ? 2 : 0);
			//save_line(buf2, nth, rc ? 2 : 0);

			fgets(buf, BUF_SIZE, fp); // "}"
			//save_line(buf, nth, rc ? 2 : 0);

			max = MAX(nth,max);
			continue;
		}

		if (type == 'h') {
			char *s, *t;

			fgets(buf1, BUF_SIZE, fp); // ">hdr"
			fgets(buf2, BUF_SIZE, fp); // ">hdr"
			type = 0;

			//save_line(buf,  nth, rc ? 2 : 0);
			//save_line(buf1, nth, rc ? 2 : 0);
			//save_line(buf2, nth, rc ? 2 : 0);

			if ((s = strchr(buf1, '"')) == NULL ||
			    (t = strchr(s+1, '"')) == NULL)
				fatalf("bad name: %s\n", buf1);
			*t = '\0';
			if (contig[nth].h1 == 0) {
			  contig[nth].h1 = copy_string(s+1);
			  unRCify(contig[nth].h1);
			}

			if ((s = strchr(buf2, '"')) == NULL ||
			    (t = strchr(s+1, '"')) == NULL)
				fatalf("bad name: %s\n", buf2);
			*t = '\0';
			if (contig[nth].name == 0) {
			  contig[nth].name = copy_string(s+1);
			  unRCify(contig[nth].name);
//fprintf(stderr, "name=%s h1=%s\n", contig[nth].name, contig[nth].h1);
			}

			fgets(buf, BUF_SIZE, fp); // "}"
			//save_line(buf, nth, rc ? 2 : 0);

			continue;
		}

		if (type == 'a') {
		    save_line(buf, nth, rc ? 3 : 1);
		}
	}
	return max;
}

int get_blastz(char *file)
{
	FILE *ap;
	char buf[BUF_SIZE];
	int n;

	ap = ckopen(file, "r");
	if ((fgets(buf, BUF_SIZE, ap)) == NULL || strcmp(buf, "#:lav\n"))
		fatal("Not an alignment file.");
	n = get_contigs(ap);
	fclose(ap);
	return n;
}

void put_hdr(int i, int rc)
{
    const Contig *const c = &contig[i];

    //printf("#:lav\n");
    printf("s {\n");
    printf("  \"%s\" %d %d %d %d\n", c->s1, c->s1lo, c->s1hi, 0, 1);
    printf("  \"%s%s\" %d %d %d %d\n", c->s2, rc?"-":"", c->s2lo, c->s2hi, rc, i);
    printf("}\n");
    printf("h {\n");
    printf("  \"%s\"\n", c->h1);
    printf("  \"%s%s\"\n", c->name, rc ? RC : "");
    printf("}\n");
}

void put_lines(Line_ptr L, int lav)
{
	int i = 0;
	if (lav)
		printf("#:lav\n");
	for ( ; L; L = L->next)
		printf("%s", L->line), ++i;
}

void emit_contig_n(int i, int rc)
{
	chkC(i);
	//fprintf(stderr, "contig=%d\n", i);
	fprintf(stdout, "#{ contig=%d\n", i);
	put_hdr(i, 0);
	put_lines(contig[i].list[rc ? 3 : 1], 0);
	fprintf(stdout, "## contig=%d\n", i);
	put_hdr(i, 1);
	put_lines(contig[i].list[rc ? 1 : 3], 0);
	fprintf(stdout, "#}\n");
}

void xxx_emit_contig_n(int i, int rc)
{
	//fprintf(stderr, "contig=%d\n", i);
	fprintf(stdout, "#{ contig=%d\n", i);
	put_lines(contig[i].list[0], 1);
	put_lines(contig[i].list[rc ? 3 : 1], 0);
	fprintf(stdout, "## contig=%d\n", i);
	put_lines(contig[i].list[2], 1);
	put_lines(contig[i].list[rc ? 1 : 3], 0);
	fprintf(stdout, "#}\n");
}


int same_string3(char *a, char *b, char *c)
{
	int r;
	char *d;

	if (a == 0) a = "";
	if (b == 0) b = "";
	if (c == 0) c = "";
	d = ckalloc(strlen(b)+strlen(c)+1); {
	  strcpy(d, b);
	  strcat(d, c);
	  r = strcmp(a, d) == 0;
	}
	ckfree(d);
	return r;
}

int emit_contig_byname(int ncontig, char *name, int rc)
{
	int found = 0;
	int i;
	for (i = 1; i <= ncontig; ++i) {
//fprintf (stderr, "checking %s ?= %s\n", name, contig[i].name);
		// treat "foo" and "foo (reverse complement)" as the same
		if (same_string3(contig[i].name, name, "") ||
		    same_string3(contig[i].name, name, RC)) {
			emit_contig_n(i, rc);
//fprintf (stderr, "...found.\n");
			found++;
		}
	}
//fprintf (stderr, "found %d {%s}\n", found, name ? name : "(null)");
	return found;
}


void out_align(char *order, int ncontig)
{
	char buf[BUF_SIZE], *s, *t;
	int rc;
	FILE *fp = ckopen(order, "r");

	//emit_contig_byname(ncontig, 0, 0);

	while (fgets(buf, BUF_SIZE, fp)) {
		int found = 0;

		if ((s = strchr(buf, '>')) == NULL)
			fatalf("no '>': %s", buf);
		if ((t = strchr(buf, '\n')) != NULL)
			*t = '\0';
		if (buf[strlen(buf)-1] == '-') {
			rc = 1;
			buf[strlen(buf)-1] = '\0';
		} else
			rc = 0;

		found = emit_contig_byname(ncontig, s, rc);
		if (!found)
			fatalf("Cannot find contig '%s'\n", s);
	}
	fclose(fp);
}

int main(int argc, char **argv)
{
	int ncontig;

	argv0 = "order_lav";
	if (argc != 3)
		fatalf("args = OandO-output blastz-output");

	ncontig = get_blastz(argv[2]);

	printf("#:lav\nd {\n  order_lav\n}\n");
	out_align(argv[1], ncontig);
	return 0;
}
