/* RCSID("$Id: refine.c,v 1.8 2005/02/15 18:45:26 rico Exp $"); */

/* refine -- P=0 (default) checks that alignments in one species don't overlap.
*  P=1 allows overlap, and stores each local alignment in its own row.
*  P=2 greedily packs non-overlapping alignments into rows.  Only P=0 works. */

#include "refine.h"

#define CAUTIOUS

#define WIDTH 10                /* Default bandwidth */
#define SUFFIX "align"
#define TBACK_SIZE 50000000
#define BUF_SIZE 5000

/* Each of the main units of sequence data is called a "piece".  P[0] is
*  simply the "reference" sequence (species 0).  Piece P[i], for 0 < i < nP,
*  is the range of the i-th alignment.
*/

/* ---------------- Read the pieces and pairwise alignments --------------- */

static int count_alignments(char *species)
{
	FILE *fp;
	char align_file[BUF_SIZE], buf[BUF_SIZE];
	int count;

	sprintf(align_file, "%s.%s", species, SUFFIX);
	fp = ckopen(align_file, "r");
	count = 0;
	while (fgets(buf, BUF_SIZE, fp))
		if (strncmp(buf, "a {", 3) == 0)
			++count;
	fclose(fp);
	if (count == 0)
		fatalf("%s contains no alignments", align_file);
	return count;
}

static align_ptr new_align_node(int k)
{
	align_ptr new;
	int i;

	if (k <= 0)
		fatal("requesting empty alignment column");
	new = ckalloc(sizeof(align_node)); 
	new->col = ckalloc(k);
	for (i = 0; i < k; ++i)
		new->col[i] = FREE;
	new->colno = -1;
	return new;
}

static void piece_info(piece_ptr p, uchar *seq, char *name, int len, int beg,
  int end, int row)
{
	p->seq = ckalloc(len+1);
	strncpy((char *)p->seq, (char *)seq, len);
	p->name = copy_string(name);
	p->len = len;
	p->beg = beg;
	p->end = end;
	p->row = row;
}

/* generate a multiple alignment containing only the first sequence */
static align_ptr init_align(char *seqname, int nRow, piece_ptr P)
{
	SEQ *sf;
	align_ptr start = NULL, end = NULL, new;
	uchar *s;
	int i;

	sf = seq_open(seqname);
	if (!seq_read(sf) || SEQ_LEN(sf) == 0)
		fatal("First sequence cannot be read.");
	piece_info(P, SEQ_CHARS(sf), seqname, SEQ_LEN(sf), 1, SEQ_LEN(sf), 0);
	if (seq_read(sf))
		fatal("First sequence contains more than one contig.");
	seq_close(sf);

	s = P[0].seq;
	for (i = 0; i < P[0].len; i++) {
		new  = new_align_node(nRow);
		new->col[0] = (char)s[i];
		new->colno = i+1;
		if (start == NULL)
			start = new;
		new->prev = end;
		if (end != NULL)
			end->next = new;
		end = new;
	}
	end->next = NULL;
	start->prev = end;
	P[0].firstcol = start;
	P[0].lastcol = end;
	return start;
}

static void store(align_ptr ap, int pos, char c, char *sname)
{
	if (ap->col[pos] != FREE) 
		fatalf("Overlapping matches in %s to pos %d in base seq.", 
			sname, ap->colno);
	ap->col[pos] = c;
}

static char *ckget(char *buf, int size, FILE *fp, char *filename)
{
	char *ret;

	if ((ret = fgets(buf, size, fp)) == NULL)
		fatalf("error reading %s", filename);
	return ret;
}

static int pairwise(align_ptr align, char *seq_name, int Pwhich,
  piece_ptr P, int nP, int nRow, int *row_ptr)
{
	piece_ptr p = &P[Pwhich];
	align_ptr new, ap = align;
	SEQ *sf = seq_open(seq_name);
	FILE *fp;
	uchar *contig = NULL;
	char buf[BUF_SIZE], buf2[BUF_SIZE], fname[BUF_SIZE],
	  *contig_name = NULL, *s, *t;
	int row = *row_ptr;
	int i, j, ncontigs, b1, b2, e1, e2, beg, end, len, prev_e2 = 0,
	  rc_curr, rc_lav, contig_curr, contig_lav;

	/* Throughout, ap->colno is the position (>= 1) in seq 0, else -1 */

	/* How many contigs in the sequence file? */
	fp = ckopen(seq_name, "r");
	ncontigs = 0;
	while (fgets(buf, BUF_SIZE, fp))
		if (buf[0] == '>')
			++ncontigs;
	fclose(fp);

	sprintf(fname, "%s.%s", seq_name, SUFFIX);
	fp = ckopen(fname, "r");
	ckget(buf, BUF_SIZE, fp, fname);
	if (!same_string(buf, "#:lav\n"))
		fatalf("%s is not a blastz output file.", buf2);

	rc_curr = contig_curr = 0;
  /* Read through the alignment. */
	while (fgets(buf, BUF_SIZE, fp))
    /* 1. LAV block */ if (same_string(buf, "#:lav\n")) {
		/* Get the new contig. */
		// Notice when we start on a new sequence, 
		// modulo reverse compliment, and read it
		while (ckget(buf, BUF_SIZE, fp, fname) && strcmp(buf, "s {\n"))
			;
		ckget(buf, BUF_SIZE, fp, fname);
		ckget(buf, BUF_SIZE, fp, fname);
		if (3 != sscanf(buf, " %s %*d %*d %d %d", buf2,
		  &rc_lav, &contig_lav))
			fatalf("bad sequence record {%s}", buf);
		if (contig_lav < contig_curr)
			fatal("impossibly small contig number");
		while (contig_curr < contig_lav) {
			if (!seq_read(sf))
				fatalf("Not enough contigs in %s.", seq_name);
			++contig_curr;
			rc_curr = 0;
		}
		if (rc_curr != rc_lav) {
			seq_revcomp_inplace(sf);
			rc_curr = rc_lav;
		}
		if ((s = strstr(buf2, "-\"")) != 0) {// XXX -- arggg
			*s++ = '"';
			*s++ = 0;
		}

		// find headers
		while (ckget(buf, BUF_SIZE, fp, fname) && strcmp(buf, "h {\n"))
			;
		ckget(buf, BUF_SIZE, fp, fname);
		ckget(buf, BUF_SIZE, fp, fname);
		if (strncmp(buf, "   \">", 5))
			fatalf("not a contig name: %s", buf);
		if (contig_name)
			free(contig_name);
		for (s = buf+5; *s == ' '; ++s)
			; 
		for (t = s; *t && *t != ' ' && *t != '"'; ++t)
			;
		*t = '\0';
		contig_name = copy_string(s);

		contig = SEQ_CHARS(sf);
		ap = align;
    /* 2. Alignment */ } else if (same_string(buf, "a {\n")) {
		/* Record the new piece. */
		if (Pwhich >= nP)
			fatal("Too many pieces.");
		p = &P[Pwhich++];
		ckget(buf, BUF_SIZE, fp, fname);
		ckget(buf, BUF_SIZE, fp, fname);
		if (sscanf(buf, "  b %*d %d", &beg) != 1)
			fatalf("Cannot find beginning: %s", buf);
		ckget(buf, BUF_SIZE, fp, fname);
		if (sscanf(buf, "  e %*d %d", &end) != 1)
			fatalf("Cannot find end: %s", buf);
		len = end - beg + 1;
		strcpy(buf2, seq_name);
		if (ncontigs > 1) {
			strcat(buf2, ":");
			strcat(buf2, contig_name);
		}
		for (s = buf2; *s; ++s)
			if (strchr(" \t\n", *s))
				break;
		*s = '\0';
		if (rc_curr)
			strcat(buf2, "-");

		piece_info(p, contig+beg-1, buf2, len, beg, end, row);
		prev_e2 = 0;
    /* 3. Gap-free aligned segment */	} else if
		(sscanf(buf," l %d %d %d %d",&b1,&b2,&e1,&e2) == 4) {
	/* 3a. Handle gap between segments with extra characters in the current
	    piece. */
		if (prev_e2) {
		    if (ap->colno != b1-1 && prev_e2 != b2-1)
			fatal("bad inter-segment spacing");
		    for (j = prev_e2+1; j < b2; ++j) {
			/* For each extra letter in the current piece either: */
			if ((ap->next)->col[0] == DASH) {
				/* Add entry to existing column. */
				ap = ap->next;
				store(ap, row, contig[j-1], seq_name);
			} else {
				/* Create new column. */
				new = new_align_node(nRow);
				for (i = 0; i < row; ++i)
					new->col[i] = DASH;
				store(new, row, contig[j-1], seq_name);
				new->prev = ap;
				new->next = ap->next;
				new->next->prev = new;
				ap = ap->next = new;
			}
		    }
		}
	/* 3b. Move along links to the segment's start in the P[0].seq. */
		while (ap->colno < b1) {
			if ((ap = ap->next) == NULL)
				fatalf("crash 1 pos %d", ap->colno);
			if (prev_e2 && ap->colno < b1)
				/* Extra letter in P[0].seq; put a dash in
				   the current row. */
				store(ap, row, DASH, seq_name);
		}
	/* 3c. Match positions in the gap-free segment. */
		if (prev_e2 == 0)
			p->firstcol = ap;
		j = b2 - 1;
		store(ap, row, contig[j++], seq_name);
		while (ap->colno < e1) {
			if ((ap = ap->next) == NULL)
				fatalf("crash 2 at %d", ap->colno);
			store(ap, row,
			  (ap->col[0] != DASH) ? contig[j++] : DASH,
			  seq_name);
		}
		p->lastcol = ap;
		prev_e2 = e2;
	}

	if (contig_name)
		free(contig_name);
	contig_name = NULL;
	seq_close(sf);
	fclose(fp);
	*row_ptr = row + 1;
	return Pwhich;
}

/* ----------------------- output the multiple alignment ------------------ */

static void mrow(align_ptr align, align_ptr from, align_ptr end, int row)
{
	int j, c, at_start;

	printf(" l");
	at_start = from->colno;
	j = 0;
	for ( ; from != end; from = from->next) {
		if (isalpha(c = from->col[row]))
			if (j >= 0)
				j++;
			else {
				printf(" %d", j - at_start);
				at_start = 0;
				j = 1;
			}
		else if (c == DASH || c == FREE)
			if (j <= 0)
				j--;
			else {
				if (at_start) {
					printf(" %d", -at_start);
					at_start = 0;
				}
				printf(" %d", j);
				j = -1;
			}
		else
			fatalf("bad character '%c' in row", c, row);
	}
	if (at_start)
		printf(" %d", -at_start);
	if (from != NULL) {
		if (j < 0)
			printf(" %d",
			  j -(align->prev->colno - from->colno + 1));
		else
			printf(" %d %d",
			  j, -(align->prev->colno - from->colno + 1));
	} else
		printf(" %d", j);
	putchar('\n');
}
	
static void print_malign(int argc, char **argv, align_ptr align, int score,
  piece_ptr P, int nP)
{
	int i;
	align_ptr ap;

	printf("#:lat\n\nd {\n \"");
	while (argc-- > 0) {
		printf("%s ", *argv++);
	}
	printf("\"\n}\n");

	printf("s {\n");
	for (i = 0; i < nP; i++)
		printf("\"%s\" %d %d\n", P[i].name, P[i].beg, P[i].end);
	printf("}\na {\n s %d\n b", score);
	for (i = 0; i < nP; i++) {
		printf(" %d", 1);
	}
	printf("\n e");
	for (i = 0; i < nP; i++)
		printf(" %d", P[i].len);
	printf("\n");

	for (i = 0, ap=align; ap; ap = ap->next, ++i)
		ap->colno = i;

	for (i = 0; i < nP; i++)
		mrow(align, P[i].firstcol, P[i].lastcol->next, P[i].row);
	printf("}\n");
}

static void free_all(align_ptr a, piece_ptr P, int nP)
{
	align_ptr b;
	int i;

	for ( ; a; a = b) {
		b = a->next;
		free(a->col);
		free(a);
	}
	for (i = 0; i < nP; ++i) {
		free(P[i].seq);
		free(P[i].name);
	}
	free(P);
}

#ifdef CAUTIOUS
static int bad_pair(align_ptr a, int nRow)
{
	align_ptr b;
	int r;
	uchar x, y;

	for ( ; (b = a->next) != NULL; a = b) 
		for (r = 0; r < nRow; ++r) {
			x = a->col[r];
			y = b->col[r];
			if ((x == DASH && y == FREE) ||
			    (x == FREE && y == DASH) ) {
				fprintf(stderr,
				  "x = %c, y = %c, col %d, row %d\n",
				   x, y, a->colno, r); 
				return 1;
			}
		}
	return 0;
}

static int bad_free(align_ptr A, piece_ptr P, int nP)
{
	align_ptr a;
	int p, r, cn, c;

	for (cn = 1, a = A; a; ++cn, a = a->next)
		a->colno = cn;
	for (p = 0; p < nP; ++p) {
		r = P[p].row;
		for (a = P[p].firstcol; a != P[p].lastcol->next; a = a->next)
			if ((c = a->col[r]) != DASH && !isalpha(c)) {
				fprintf(stderr, 
				  "Unwanted character `%c' in alignment,", c);
				fprintf(stderr,
				  "  piece %d, row %d, column %d\n",
				  p, r, a->colno);
				return 1;
			}
	}
	return 0;
}

/* bad_nonfree checks that only FREE occurs between pieces in the alignment;
*   the alignment's entries are distroyed in the process */
static int bad_nonfree(align_ptr A, int nRow, piece_ptr P, int nP)
{
	align_ptr a;
	int p, r, cn;
	uchar *acol;

	for (cn = 1, a = A; a; ++cn, a = a->next)
		a->colno = cn;
	for (p = 0; p < nP; ++p) {
		r = P[p].row;
		for (a = P[p].firstcol; a != P[p].lastcol->next; a = a->next)
			a->col[r] = FREE;
	}
	for (a = A; a; a = a->next) {
		acol = a->col;
		for (r = 0; r < nRow; ++r)
			if (acol[r] != FREE) {
				fprintf(stderr,
				  "`%c' between pieces at row %d, col %d\n",
				  acol[r], r, a->colno);
				return 1;
			}
	}
	return 0;
}
#endif

#define DEFAULT_ITER 1000
#define DEFAULT_DEPTH 10
int main(int argc, char *argv[])
{
	align_ptr align, ap;
	char *scorefile;
	gap_scores_t ds;
	ss_t ss;
	int gop[256], gtype[128];
	int flag_overlap, Pwhich, nRow, row, score, i, nS, width, nP,
	  new_score, flag_iter, flag_depth, r, p;
	piece_ptr P;
	TBack tback;

	argv0 = "refine";
	if (argc < 2)
		fatal("seq1 seq2 .. seqN [w=<width> [Q=scores] [PTD=?]");

	/* Determine number of species and of rows. */
	for (nS = argc-1; strchr(argv[nS], '='); --nS) 
		;
	for (nP = 1, i = 2; i <= nS; ++i)
		nP += count_alignments(argv[i]);

	ckargs("QwPTD", argc, argv, nS+1);
	get_argval_pos('w', &width, WIDTH);
	get_argval_nonneg('P', &flag_overlap, 0);
	get_argval_nonneg('T', &flag_iter, DEFAULT_ITER);
	get_argval_nonneg('D', &flag_depth, DEFAULT_DEPTH);
	if (flag_overlap > 0) fatal("positive overlap flag not implemented");
	nRow = (flag_overlap ? nP : nS);	/* less if flag_overlap == 1 */

	/* Set alignment scores. */
	if (get_cargval('Q', &scorefile))
		scores_from_file(scorefile, ss);
	else
		DNA_scores(ss);
	ds.E = 30;
	ds.O = 400;
	for (i = 0; i < 128; ++i)
		ss[i][DASH] = ss[DASH][i] = -ds.E;
	ss[DASH][DASH] = 0;
	for (i = 0; i < 128; ++i)
		ss[i][FREE] = ss[FREE][i] = 0;
	gap_costs(gop, gtype, ds.O);

	P = (piece_ptr)ckalloc(nP*sizeof(piece));

	/* Create an alignment containing only sequence 0. */
	align = init_align(argv[1], nRow, P);
	
	/* Build the crude multiple alignment from the pairwise alignments,
	*  determine the Pieces, and (when flag_overlap == 2) re-set nRow.
	*/
	for (row = 1, Pwhich = i = 1; i < nS; ++i)
		Pwhich = pairwise(align, argv[i+1], Pwhich, P, nP, nRow,
		  &row);
	nRow = row;
#ifdef CAUTIOUS
	if (bad_free(align, P, nP))
		fatal("Bad FREE in very initial alignment.");
#endif

	for (ap = align; ap; ap = ap->next)
		for (r = 0; r < nRow; ++r)
			if (ap->col[r] == DASH)
				ap->col[r] = FREE;
	for (p = 0; p < nP; ++p) {
		r = P[p].row;
		for (ap = P[p].firstcol; ap!=P[p].lastcol->next; ap = ap->next)
			if (ap->col[r] == FREE)
				ap->col[r] = DASH;
		}

#ifdef CAUTIOUS
	if (bad_free(align, P, nP))
		fatal("Bad FREE in initial alignment.");
	if (bad_pair(align, nRow))
		fatal("Bad FREE/DASH pair in initial alignment");
#endif
	score = align_score(align, align->prev, nRow, ss, gop, gtype);


	if (flag_iter) {
		tback = ckalloc(sizeof(TB));
		tback->space = ckalloc(TBACK_SIZE*sizeof(uchar));
		tback->len = TBACK_SIZE;
		align = robin(align, nRow, P, nP, width, ss, gop, gtype,
		   flag_iter, flag_depth, tback);
		new_score = align_score(align, align->prev, nRow, ss, gop,
		  gtype);
		fprintf(stderr, "Score changed from %d to %d.\n",
		  score, new_score);
		score = new_score;
		free(tback->space);
		free(tback);
	}

	print_malign(argc, argv, align, score, P, nP);

#ifdef CAUTIOUS
	if (bad_pair(align, nRow))
		fatal("Bad FREE/DASH pair in final alignment");
	if (bad_free(align, P, nP))
		fatal("Bad FREE in final alignment.");
	if (bad_nonfree(align, nRow, P, nP))
		fatal("Bad non-FREE in final alignment.");
#endif

	free_all(align, P, nP);


	return 0;
}
